import { SchemaReactflowNodeWithFunctions, SchemaReactflowRelation, SchemaReactflowRelationWithFunctions } from '../model/reactflow';
import Graph from 'graphology';
import { Attributes } from 'graphology-types';
import { MarkerType, Edge, Node } from 'reactflow';
import { SchemaReactflowNode } from '../model/reactflow';
import { QueryElementTypes } from '../../querybuilder';

//TODO does not belong here; maybe should go into the GraphPolarisThemeProvider
const ANIMATEDEDGES = false;

export function schemaExpandRelation(graph: Graph): Graph {
  const newGraph = graph.copy();

  newGraph.forEachNode((node, attributes) => {
    // console.log(node, attributes);
    newGraph.mergeNodeAttributes(node, { type: QueryElementTypes.Entity });
  });

  //makeNewRelationNodes
  graph.forEachEdge((edge, attributes, source, target): void => {
    const newID = 'RelationNode:' + edge;
    // console.log('making relationnode', edge, attributes, source, target, newID);
    newGraph.addNode(newID, {
      name: edge,
      label: edge,
      // data: {
      //   label: edge,
      //   name: edge,
      // attributes: attributes
      // },
      ...attributes,
      x: 0,
      y: 0,
      type: QueryElementTypes.Relation,
    });

    const id = 'RelationEdge' + source + '->' + newID;
    newGraph.addDirectedEdgeWithKey(id, source, newID, {
      name: edge,
    });

    const id2 = 'RelationEdge' + newID + '->' + target;
    // console.log('making newEdgeOutgoing', edge, attributes, source, target, id);
    newGraph.addDirectedEdgeWithKey(id2, newID, target, {});

    newGraph.dropEdge(edge);
  });

  return newGraph;
}

// Takes the schema as an input and creates basic react flow elements for them.
export function schemaGraphology2Reactflow(graph: Graph): {
  nodes: Array<Node<SchemaReactflowNodeWithFunctions | SchemaReactflowRelationWithFunctions>>;
  edges: Array<Edge>;
} {
  const initialElements: { nodes: Array<Node>; edges: Array<Edge> } = {
    nodes: [],
    edges: [],
  };

  initialElements.nodes = createReactFlowNodes(graph);
  initialElements.edges = createReactFlowEdges(graph);
  // initialElements.push(...createReactFlowRelationNodes(graph));
  // initialElements.push(...createReactFlowRelationEdges(graph));
  // console.log(initialElements);

  return initialElements;
}

export function createReactFlowNodes(graph: Graph): Array<Node> {
  const nodeElements: Array<Node> = [];
  graph.forEachNode((node: string, attributes: Attributes): void => {
    const newNode: Node = {
      id: node,
      data: {
        ...attributes,
      },
      position: { x: attributes.x, y: attributes.y },
      type: attributes.type,
    };
    nodeElements.push(newNode);
  });

  return nodeElements;
}

export function createReactFlowEdges(graph: Graph): Array<Edge> {
  const edgeElements: Array<Edge> = [];

  graph.forEachEdge((edge, attributes, source, target): void => {
    // console.log('edges', attributes, edge);
    const newEdge: Edge = {
      id: edge,
      source: source,
      target: target,
      data: {
        ...attributes,
      },
      // label: edge,
      type: attributes?.type || 'smoothstep',
      animated: ANIMATEDEDGES,
      markerEnd: MarkerType.ArrowClosed, // TODO: Check
    };
    edgeElements.push(newEdge);
  });

  return edgeElements;
}

// export function createReactFlowRelationNodes(graph: Graph): Elements<Node> {
//   const nodeElements: Elements<Node> = [];
//   graph.forEachEdge((edge, attributes, source, target): void => {
//     const newRelationNode: Node = {
//       id: edge,
//       data: {
//         label: edge,
//         name: edge,
//       },
//       position: { x: attributes.x, y: attributes.y },
//       type: QueryElementTypes.Relation,
//     };
//     nodeElements.push(newRelationNode);
//   });

//   return nodeElements;
// }

// export function createReactFlowRelationEdges(graph: Graph): Elements<Edge> {
//   const edgeElements: Elements<Edge> = [];
//   graph.forEachEdge((edge, attributes, source, target): void => {
//     const newEdgeIncoming: Edge = {
//       //into relation node
//       id: edge + '' + source,
//       source: source,
//       target: edge,
//       // label: edge,
//       type: 'smoothstep',
//       animated: ANIMATEDEDGES,
//       arrowHeadType: ArrowHeadType.ArrowClosed,
//     };
//     edgeElements.push(newEdgeIncoming);

//     const newEdgeOutgoing: Edge = {
//       //out of relation node
//       id: edge + '' + target,
//       source: edge,
//       target: target,
//       // label: edge,
//       type: 'smoothstep',
//       animated: ANIMATEDEDGES,
//       arrowHeadType: ArrowHeadType.ArrowClosed,
//     };
//     edgeElements.push(newEdgeOutgoing);
//   });

//   return edgeElements;
// }

// export function parseSchemaFromBackend(
//   schemaFromBackend: SchemaFromBackend
// ): Graph {
//   const { nodes, edges } = schemaFromBackend;
//   // Instantiate a directed graph that allows self loops and parallel edges
//   const schemaGraph = new MultiGraph({ allowSelfLoops: true });
//   // console.log('parsing schema');
//   // The graph schema needs a node for each node AND edge. These need then be connected

//   nodes.forEach((node) => {
//     schemaGraph.addNode(node.name, {
//       name: node.name,
//       attributes: node.attributes,
//       x: 0,
//       y: 0,
//     });
//   });

//   // The name of the edge will be name + from + to, since edge names are not unique
//   edges.forEach((edge) => {
//     const edgeID = [edge.name, '_', edge.from, edge.to].join(''); //ensure that all interpreted as string

//     // This node is the actual edge
//     schemaGraph.addDirectedEdgeWithKey(edgeID, edge.from, edge.to, {
//       attribute: edge.attributes,
//     });
//   });
//   return schemaGraph;
// }
